/**
 * Unboxed primitive types.
 *
 * We use boxed primitive types where possible, to achieve nominal typing for extra type safety. However, TSOA does not
 * recognise these, so we provide unboxed variants here that we can attach OpenAPI specification metadata to.
 */
export declare type IgnoreMeUnboxed = never;
/**
 * Name of a multipart/form-data field.
 * @example "file"
 * @minLength 1
 * @pattern ^.*$
 */
export declare type MultipartFormDataFieldName = string;
/**
 * Upload user ID.
 * @example "BC8048E8-A6F4-40CD-BFA2-1081049AA1AB"
 * @minLength 36
 * @maxLength 36
 * @pattern ^\S(.*?\S)?$
 */
export declare type UserIdUnboxed = string;
/**
 * Your account ID.
 *
 * This is visible on the settings page:
 *
 * https://upload.io/dashboard/settings
 * @example "A623uY2"
 * @minLength 7
 * @maxLength 7
 * @pattern ^[1-9A-HJ-NP-Za-km-z]*$
 */
export declare type AccountIdUnboxed = string;
/**
 * Absolute path to a folder. Must begin with a `/`. Should not end with a `/`.
 * @example "/uploads"
 * @minLength 1
 * @maxLength 512
 * @pattern ^/.*$
 */
export declare type FolderPathUnboxed = string;
/**
 * Absolute path to a file. Must begin with a `/`.
 * @example "/uploads/image.jpg"
 * @minLength 1
 * @maxLength 512
 * @pattern ^/.*$
 */
export declare type FilePathUnboxed = string;
/**
 * Absolute path to a file.
 * @example "uploads/image.jpg"
 * @minLength 1
 * @maxLength 512
 * @pattern ^.*$
 */
export declare type FilePathTailUnboxed = string;
/**
 * Absolute path to a file. Must begin with a `/`.
 * @example "/uploads/file.txt"
 * @minLength 1
 * @maxLength 512
 * @pattern ^/.*$
 */
export declare type FilePathTextUnboxed = string;
/**
 * Absolute or relative path to a location in your Upload account's storage.
 * @example "/uploads/image.jpg"
 * @minLength 0
 * @maxLength 512
 * @pattern ^.*$
 */
export declare type RelativeOrAbsoluteFilePathUnboxed = string;
/**
 * Absolute or relative path to a location in your Upload account's storage.
 * @example "/uploads"
 * @minLength 0
 * @maxLength 512
 * @pattern ^.*$
 */
export declare type RelativeOrAbsoluteFolderPathUnboxed = string;
/**
 * Absolute path to a file or folder. Must begin with a `/`.
 * @example "/uploads/file.jpg"
 * @minLength 1
 * @maxLength 512
 * @pattern ^/.*$
 */
export declare type FileOrFolderPathUnboxed = string;
/**
 * File name. May be empty string `""`. Cannot contain `/`.
 * @example "image.jpg"
 * @minLength 0
 * @maxLength 512
 * @pattern ^[^/]*$
 */
export declare type FileNameUnboxed = string;
/**
 * Nonce for the subscription the user is about to create.
 * @example "D812fa81A623uY2RXnNq1vfw13A63uYeFkD812fDAuY28eaFk8"
 * @minLength 50
 * @maxLength 50
 * @pattern ^[1-9A-HJ-NP-Za-km-z]*$
 */
export declare type SubscriptionNonceUnboxed = string;
/**
 * Nonce for the subscription the user is about to create.
 * @example 86400
 * @isInt
 * @minimum 0
 * @maximum 315360000
 */
export declare type TTLSecondsUnboxed = number;
/**
 * The ID for the multipart file upload.
 * @example "Kd759aLFxttm69kZ"
 * @minLength 16
 * @maxLength 16
 * @pattern ^[1-9A-HJ-NP-Za-km-z]*$
 */
export declare type UploadIdUnboxed = string;
/**
 * File uniquifier.
 * @example "Kd759aLFxttm69kZ"
 * @minLength 16
 * @maxLength 16
 * @pattern ^[1-9A-HJ-NP-Za-km-z]*$
 */
export declare type FileUniquifierUnboxed = string;
/**
 * File ID.
 * @example "W142hJkKd759aLFxttm69kZ"
 * @minLength 23
 * @maxLength 23
 * @pattern ^[1-9A-HJ-NP-Za-km-z]*$
 */
export declare type FileIdUnboxed = string;
/**
 * Size in bytes (big).
 * @example "856974744186387529544875589443785995421"
 * @minLength 1
 * @maxLength 100
 * @pattern ^\d*$
 */
export declare type BigSizeInBytesUnboxed = string;
/**
 * The tags you specify here determine which rules will be executed for the upload.
 *
 * You create rules in the Upload Dashboard, and choose which tags trigger each rule.
 *
 * Rules include: max file size checks, traffic limit checks, rate limit checks, etc.
 * @example "images/profile"
 * @minLength 1
 * @maxLength 255
 * @pattern ^[a-zA-Z0-9\.,/=_+-]*$
 */
export declare type FileTagNameUnboxed = string;
/**
 * If true performs a prefix search using the fileTag, else performs an exact-match search.
 * @example true
 */
export declare type FileTagPrefixSearchUnboxed = boolean;
/**
 * File MIME type.
 * @example "image/jpeg"
 * @minLength 3
 * @maxLength 255
 * @pattern ^[a-z0-9]+/[a-z0-9+\-._]+(?:;[^=]+=[^;]+)*$
 */
export declare type MimeTypeUnboxed = string;
/**
 * The file's original name on the user's device.
 * @example "example.jpg"
 * @minLength 1
 * @maxLength 255
 * @pattern ^.+$
 */
export declare type OriginalFileNameUnboxed = string;
/**
 * File ETag.
 * @example "33a64df551425fcc55e4d42a148795d9f25f89d4"
 * @minLength 1
 * @maxLength 255
 * @pattern ^\S*$
 */
export declare type ETagUnboxed = string;
/**
 * AWS Access Key.
 * @example "AKIAIOSFODNN7EXAMPLE"
 * @minLength 20
 * @maxLength 20
 * @pattern ^\S*$
 */
export declare type AwsAccessKeyUnboxed = string;
/**
 * AWS Secret Access Key.
 * @example "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY"
 * @minLength 40
 * @maxLength 40
 * @pattern ^\S*$
 */
export declare type AwsSecretKeyUnboxed = string;
/**
 * AWS S3 Bucket Name.
 * @example "my-bucket"
 * @minLength 1
 * @maxLength 61
 * @pattern (?!(^xn--|-s3alias$))^[a-z0-9][a-z0-9-]{1,61}[a-z0-9]$
 */
export declare type S3BucketNameUnboxed = string;
/**
 * AWS S3 Object Key.
 * @example "an/example/key"
 * @minLength 0
 * @maxLength 1024
 * @pattern ^\S*$
 */
export declare type S3ObjectKeyUnboxed = string;
/**
 * AWS Region.
 * @example "us-east-1"
 * @minLength 1
 * @maxLength 20
 * @pattern ^[a-z0-9-]*$
 */
export declare type AwsRegionUnboxed = "us-east-2" | "us-east-1" | "us-west-1" | "us-west-2" | "af-south-1" | "ap-east-1" | "ap-southeast-3" | "ap-south-1" | "ap-northeast-3" | "ap-northeast-2" | "ap-southeast-1" | "ap-southeast-2" | "ap-northeast-1" | "ca-central-1" | "eu-central-1" | "eu-west-1" | "eu-west-2" | "eu-south-1" | "eu-west-3" | "eu-north-1" | "me-south-1" | "sa-east-1";
/**
 * The number of parts the file will be uploaded with.
 * @example 12
 * @isInt
 * @minimum 1
 */
export declare type FilePartCountUnboxed = number;
/**
 * Index of an uploadable file part.
 *
 * Can be used as the `uploadPartIndex` parameter in the GetUploadPart and CompleteUploadPart endpoints.
 * @example 7
 * @isInt
 * @minimum 0
 */
export declare type UploadPartIndexUnboxed = number;
/**
 * Pre-signed upload URL for this part.
 *
 * You are required to issue a `PUT` to this URL, with the file's bytes as the request body (limited to the range indicated by this upload part).
 *
 * The `PUT` request will return an `etag` response header, which must be provided in a subsequent CompleteUploadPart request.
 * @example "https://...long-url...x-id=PutObject"
 * @minLength 10
 */
export declare type PreSignedUploadUrlUnboxed = string;
/**
 * Folder description.
 * @example "This is an example folder description."
 * @minLength 1
 * @maxLength 100
 * @pattern ^\S(.*?\S)?$
 */
export declare type FolderDescriptionUnboxed = string;
/**
 * URL for an http(s) resource.
 * @example "https://example.com/foo"
 * @minLength 1
 * @maxLength 255
 * @pattern ^https?://(www\.)?[-a-zA-Z0-9\x40:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()\x40:%_\+.~#?&//=*]*)$
 */
export declare type HttpUrlUnboxed = string;
/**
 * URL for a raw file hosted on the Upload CDN.
 * @example "https://upcdn.io/A623uY2/raw/uploads/image.jpg"
 * @minLength 1
 * @maxLength 255
 * @pattern ^https?://(www\.)?[-a-zA-Z0-9\x40:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()\x40:%_\+.~#?&//=*]*)$
 */
export declare type RawFileUrlUnboxed = string;
/**
 * URL for a raw file hosted on the Upload CDN.
 * @example "https://upcdn.io/A623uY2/raw/uploads/file.txt"
 * @minLength 1
 * @maxLength 255
 * @pattern ^https?://(www\.)?[-a-zA-Z0-9\x40:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()\x40:%_\+.~#?&//=*]*)$
 */
export declare type RawFileUrlTextUnboxed = string;
/**
 * If `true` then path variables like `{DATE_UTC}` in the `fileName` will be replaced. You can escape `{` characters with a `\`.
 *
 * If `false` then path variables like `{DATE_UTC}` in the `fileName` will be taken literally.
 *
 * Default: true
 * @example true
 */
export declare type FileNameVariablesEnabledUnboxed = boolean;
/**
 * If `true` then path variables like `{DATE_UTC}` in the `folderPath` will be replaced. You can escape `{` characters with a `\`.
 *
 * If `false` then path variables like `{DATE_UTC}` in the `folderPath` will be taken literally.
 *
 * Default: true
 * @example true
 */
export declare type FolderPathVariablesEnabledUnboxed = boolean;
/**
 * Domain name.
 * @example "localhost"
 * @minLength 1
 * @maxLength 255
 * @pattern ^[a-z0-9-\.]*$
 */
export declare type DomainNameUnboxed = string;
/**
 * Custom domain name.
 * @example "acme.com"
 * @minLength 1
 * @maxLength 255
 * @pattern ^([a-z0-9-]+\.)+[a-z]{2,}$
 */
export declare type CustomDomainNameUnboxed = string;
/**
 * Custom domain name target.
 * @example "xyz.upcdn.io"
 * @minLength 1
 * @maxLength 255
 * @pattern ^([a-z0-9-]+\.)+[a-z]{2,}$
 */
export declare type CustomDomainNameTargetDomainUnboxed = string;
/**
 * JSON Web Token (JWT).
 * @example "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiJ4In0.fiX9xWK-9af81KNXzUT_PARqLr9S43OS5iZI7MvLAYk"
 * @pattern ^[a-zA-Z0-9\-_]+?\.[a-zA-Z0-9\-_]+?\.([a-zA-Z0-9\-_]+)?$
 */
export declare type JwtUnboxed = string;
/**
 * Paddle plan ID.
 * @example 742238
 * @isInt
 * @minimum 0
 */
export declare type PaddlePlanIdUnboxed = number;
/**
 * Path to an artifact generated by the transformation.
 *
 * All image transformations output the transformed image to the `"/"` artifact. Video transformations may output multiple artifacts.
 *
 * Default: "/"
 * @example "/example/video/part-a.ts"
 * @pattern ^/(?:((?!\.*[/\\]).)((?![/\\]\.*[/\\]).)*)?$
 */
export declare type TransformationArtifactPathUnboxed = string;
/**
 * First name of a person.
 * @example "Mark"
 * @minLength 1
 * @maxLength 35
 * @pattern ^\S(.*?\S)?$
 */
export declare type FirstNameUnboxed = string;
/**
 * Last name of a person.
 * @example "Smith"
 * @minLength 1
 * @maxLength 35
 * @pattern ^\S(.*?\S)?$
 */
export declare type LastNameUnboxed = string;
/**
 * Email address.
 * @example "mark.smith\u0040example.com"
 * @minLength 1
 * @maxLength 254
 * @pattern ^\S+\x40\S+\.\S+$
 */
export declare type EmailAddressUnboxed = string;
/**
 * Organization name.
 * @example "Acme Corp"
 * @minLength 1
 * @maxLength 100
 * @pattern ^\S(.*?\S)?$
 */
export declare type OrganizationNameUnboxed = string;
/**
 * Department name.
 * @example "Software Development"
 * @minLength 1
 * @maxLength 100
 * @pattern ^\S(.*?\S)?$
 */
export declare type DepartmentNameUnboxed = string;
/**
 * Seconds.
 * @example 60
 * @isInt
 * @minimum 0
 */
export declare type SecondsUnboxed = number;
/**
 * Epoch milliseconds (since midnight 1 January 1970, UTC).
 * @example 1615680311115
 * @isInt
 * @minimum 0
 */
export declare type EpochMillisUnboxed = number;
/**
 * Seconds (since midnight 1 January 1970, UTC).
 * @example 1615680311
 * @isInt
 * @minimum 0
 */
export declare type EpochSecondsUnboxed = number;
/**
 * Description or notes about a rule.
 * @example "max_uploads_per_ip_daily"
 * @minLength 1
 * @maxLength 40
 * @pattern ^\S(.*?\S)?$
 */
export declare type RuleNameUnboxed = string;
/**
 * Job ID.
 * @example "01ARZ3NDEKTSV4RRFFQ69G5FAV"
 * @minLength 26
 * @maxLength 26
 * @pattern ^[0-9A-Z]$
 */
export declare type AccountJobIdUnboxed = string;
/**
 * Description or notes about a rule.
 * @example "Max 1000 uploads every 24 hours (per IP)"
 * @minLength 1
 * @maxLength 80
 * @pattern ^\S(.*?\S)?$
 */
export declare type RuleDescriptionUnboxed = string;
/**
 * Error to display when rule is triggered.
 * @example "Your IP has reached its daily upload limit."
 * @minLength 1
 * @maxLength 200
 * @pattern ^\S(.*?\S)?$
 */
export declare type RuleErrorMessageUnboxed = string;
/**
 * Error to display when a file has expired.
 * @example "This file has expired"
 * @minLength 1
 * @maxLength 200
 * @pattern ^\S(.*?\S)?$
 */
export declare type FileExpiredMessageUnboxed = string;
/**
 * Description or notes about an API key.
 * @example "Production API key"
 * @minLength 1
 * @maxLength 50
 * @pattern ^\S(.*?\S)?$
 */
export declare type ApiKeyDescriptionUnboxed = string;
/**
 * Public API key (can be used in frontend code).
 * @example "public_A623uY2RvnNq1vZ80fYgGyhKN0U7"
 * @minLength 35
 * @maxLength 35
 * @pattern ^public_[1-9A-HJ-NP-Za-km-z]*$
 */
export declare type PublicApiKeyUnboxed = string;
/**
 * Secret API key (do not use in frontend code).
 * @example "secret_A623uY2RvkUy789gT8jHamNd89qa"
 * @minLength 35
 * @maxLength 35
 * @pattern ^secret_[1-9A-HJ-NP-Za-km-z]*$
 */
export declare type SecretApiKeyUnboxed = string;
/**
 * API key.
 * @example "secret_A623uY2RvkUy789gT8jHaPLmNd8a"
 * @minLength 35
 * @maxLength 35
 * @pattern ^(public|secret)_[1-9A-HJ-NP-Za-km-z]*$
 */
export declare type ApiKeyUnboxed = string;
/**
 * API key type.
 * @example "secret"
 * @pattern ^(public|secret)$
 */
export declare type ApiKeyTypeUnboxed = "public" | "secret";
/**
 * Username of package maintainer on NPM.
 * @example "upload-io"
 * @minLength 1
 * @maxLength 50
 * @pattern ^[0-9A-Za-z-_.]*$
 */
export declare type NpmUsernameUnboxed = string;
/**
 * NPM package name.
 * @example "upload-compression-plugin"
 * @minLength 1
 * @maxLength 100
 * @pattern ^[0-9A-Za-z-_.\x40\/]*$
 */
export declare type NpmPackageNameUnboxed = string;
/**
 * Semantic version.
 * @example "1.2.0"
 * @minLength 1
 * @maxLength 50
 * @pattern ^([0-9]+)\.([0-9]+)\.([0-9]+)(?:-([0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*))?(?:\+[0-9A-Za-z-]+)?$
 */
export declare type SemanticVersionUnboxed = string;
/**
 * Description for an Upload Plugin.
 * @example "Archives, compresses and decompresses files."
 * @minLength 1
 * @maxLength 80
 * @pattern ^\S(.*?\S)?$
 */
export declare type PluginDescriptionUnboxed = string;
/**
 * Category for a transformation template.
 * @example "Image"
 * @minLength 1
 * @maxLength 30
 * @pattern ^\S(.*?\S)?$
 */
export declare type TemplateCategoryUnboxed = string;
/**
 * Codename for a template that is unique within the scope of a specific plugin version.
 * @example "compress"
 * @minLength 1
 * @maxLength 100
 * @pattern ^.*?$
 */
export declare type TemplateKeyUnboxed = string;
/**
 * Name for a transformation template.
 * @example "Compress a file"
 * @minLength 1
 * @maxLength 30
 * @pattern ^\S(.*?\S)?$
 */
export declare type TemplateNameUnboxed = string;
/**
 * Description for a transformation template.
 * @example "Compresses the file into a ZIP at a given compression level."
 * @minLength 1
 * @maxLength 80
 * @pattern ^\S(.*?\S)?$
 */
export declare type TemplateDescriptionUnboxed = string;
/**
 * Label for a transformation template parameter.
 * @example "Compression format"
 * @minLength 1
 * @maxLength 30
 * @pattern ^\S(.*?\S)?$
 */
export declare type TemplateParamLabelUnboxed = string;
/**
 * Transformation definition ID. System transformations begin with a "S-" prefix.
 * @example "F91kfA1g"
 * @minLength 8
 * @maxLength 8
 * @pattern ^(?:S-)?[1-9A-HJ-NP-Za-km-z]*$
 */
export declare type TransformationDefinitionIdUnboxed = string;
/**
 * API key ID.
 * @example "F91kfA1g"
 * @minLength 8
 * @maxLength 8
 * @pattern ^[1-9A-HJ-NP-Za-km-z]*$
 */
export declare type ApiKeyIdUnboxed = string;
/**
 * Public key ID.
 * @example "F91kfA1g"
 * @minLength 8
 * @maxLength 8
 * @pattern ^[1-9A-HJ-NP-Za-km-z]*$
 */
export declare type PublicKeyIdUnboxed = string;
/**
 * Name given to a public key.
 * @example "Public Key (www.upload.io)"
 * @minLength 1
 * @maxLength 40
 * @pattern ^\S(.*?\S)?$
 */
export declare type PublicKeyNameUnboxed = string;
/**
 * Public key certificate. Keys up to 4096-bit are supported.
 * @example "-----BEGIN PUBLIC KEY-----\u000A...\u000A-----END PUBLIC KEY-----"
 * @minLength 1
 * @maxLength 800
 * @pattern ^-----BEGIN .*?PUBLIC KEY-----.*?-----END .*?PUBLIC KEY-----$
 */
export declare type PublicKeyPEMUnboxed = string;
/**
 * Cursor (aka continuation token) for listing folder children.
 * @example "aGVsbG8="
 * @pattern ^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$
 */
export declare type ListFolderChildrenCursorUnboxed = string;
/**
 * Base64-encoded string.
 * @example "aGVsbG8="
 * @pattern ^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$
 */
export declare type Base64Unboxed = string;
/**
 * Base58-encoded string.
 * @example "aGVsbG8k"
 * @pattern ^[1-9A-HJ-NP-Za-km-z]*$
 */
export declare type Base58Unboxed = string;
/**
 * Description given to a transformation.
 * @example "Compresses the file into a ZIP at level 9 compression."
 * @minLength 1
 * @maxLength 80
 * @pattern ^\S(.*?\S)?$
 */
export declare type TransformationDescriptionUnboxed = string;
/**
 * MIME type scope, e.g. for a transformation definition.
 * @example "*"
 * @minLength 1
 * @maxLength 255
 * @pattern ^(?:\*|[a-z0-9]+\/(?:[a-z0-9+\-._]+|[a-z0-9+\-._]*\*))$
 */
export declare type MimeScopeUnboxed = string;
/**
 * File tag scope, e.g. for a transformation definition, a CDN download access token, or an API key's tag whitelist.
 * @example "*"
 * @minLength 1
 * @maxLength 255
 * @pattern ^[a-zA-Z0-9\.,/=_+-]*\*?$
 */
export declare type TagScopeUnboxed = string;
/**
 * Identifies the location to save/load transformed files to/from.
 * @example "compressed-files-v1"
 * @minLength 8
 * @maxLength 8
 * @pattern ^[a-zA-Z0-9-]*$
 */
export declare type CacheIdUnboxed = string;
/**
 * Transformation URL slug.
 *
 * These are created in the Upload Dashboard:
 *
 * https://upload.io/dashboard/transformations
 * @example "thumbnail"
 * @minLength 1
 * @maxLength 100
 * @pattern ^[a-zA-Z0-9-_+=\x40~;%][a-zA-Z0-9-_.+=\x40~;%]*$
 */
export declare type TransformationUrlSlugUnboxed = string;
/**
 * Transformation URL slug.
 *
 * - Use `"*"` to allow all file downloads.
 *
 * - Use `"raw"` to allow raw/original file downloads.
 *
 * - Use a `*` suffix to allow transformation prefixes (e.g. `"thumbnail-*"` will allow `thumbnail-sm` and `thumbnail-lg`).
 *
 * - Use any other value to allow specific transformations (e.g. `"thumbnail"` will allow `thumbnail` downloads only).
 *
 * @example "thumbnail"
 * @minLength 1
 * @maxLength 100
 * @pattern ^[a-zA-Z0-9-_.+=\x40~;%]*\*?$
 */
export declare type TransformationUrlSlugPatternUnboxed = string;
/**
 * @minItems 1
 */
export declare type NonEmptyArrayUnboxed<T> = T[];
/**
 * @uniqueItems true
 * @minItems 1
 */
export declare type NonEmptySetUnboxed<T> = T[];
/**
 * Continuation token (aka pagination cursor).
 * @example "W142hJkKd759aLFxttm69kZ..."
 * @minLength 1
 * @maxLength 1000
 * @pattern ^\S(.*?\S)?$
 */
export declare type ContinuationTokenUnboxed = string;
/**
 * Page size (i.e. maximum items to return in a page of results).
 * @example 12
 * @isInt
 * @minimum 1
 * @maximum 250
 */
export declare type FileApiV1PageSizeUnboxed = number;
/**
 * Folder listing limit.
 * @example 50
 * @isInt
 * @minimum 1
 * @maximum 200
 */
export declare type ListFolderChildrenLimitUnboxed = number;
/**
 * Non-negative integer.
 * @example 42
 * @isInt
 * @minimum 0
 */
export declare type NonNegativeIntUnboxed = number;
/**
 * Account member role.
 */
export declare type AccountMemberRoleUnboxed = "AccountAdmin" | "AccountObserver";
