/**
 * Utility method for working with whitelists defined as arrays of strings, where each string is either an absolute
 * matcher, or a prefix matcher (if it has a '*' suffix) or a wildcard matcher (if is a '*' character).
 */
export class WhitelistUtils {
  /**
   * Each element in 'pattern' is combined through an OR relationship.
   *
   * Each element in "pattern" is either an:
   * - Absolute (e.g. "images/jpeg")
   * - Prefix (e.g. "images/*")
   * - Wildcard (i.e. '*')
   * Element element in "value" is absolute.
   */
  static intersect(whitelist: string[], values: string[]): boolean {
    const isMatch = (pattern: string, value: string[]): boolean =>
      pattern === "*" || value.some(v => (pattern.endsWith("*") ? v.startsWith(pattern.slice(0, -1)) : pattern === v));

    return whitelist.some(x => isMatch(x, values));
  }

  static includes(whitelist: string[], value: string): boolean {
    return whitelist.some(x => x === "*" || (x.endsWith("*") ? value.startsWith(x.slice(0, -1)) : x === value));
  }

  static intersectOrEmpty(whitelist: string[], values: string[]): boolean {
    if (values.length === 0) {
      // Required for when users upload files with no tags: they must pass tag whitelists on API keys. Whitelists
      // are intended to restrict what users can do, not force them to do more (i.e. by requiring them to add tags).
      // Also: an empty whitelist means "you cannot upload to any tags", which is valid, but without returning true
      // below, this would mean all uploads may fail by empty whitelists, rendering empty whitelists useless, and would
      // effectively behave as "disable all uploads"... so what's the point in an API key that disables all uploads?
      // If the user specifically wants to disable uploads with no tags, then they should create a 'reject rule' that
      // matches on the 'unassigned' tag. :)
      return true;
    }

    return WhitelistUtils.intersect(whitelist, values);
  }
}
