import { groupBy, mapObjIndexed } from "ramda";
import { DeepUndefinedToNull, UnboxedDto } from "@upload-io/upload-shared-iso";

export class IsoObjectUtils {
  static box<TBoxedDto, TUnboxedDto extends UnboxedDto<TBoxedDto> = UnboxedDto<TBoxedDto>>(
    unboxedDto: TUnboxedDto
  ): TBoxedDto {
    // Given the unboxed DTO is assignable to UnboxedDto<TBoxedDto>, then the only runtime change required to
    // fit UnboxedDto<TBoxedDto> into TBoxedDto is to replace all 'null' values with 'undefined': the rest is just
    // coercing Array<T> instances into NonEmptyArray<T>, and T into Opaque<T, any>. We assume this is safe, as the
    // unboxed definitions _should_ have OpenAPI Spec annotations ensuring things like cardinality and syntax.
    return IsoObjectUtils.nullToUndefined(unboxedDto) as any;
  }

  static nullToUndefined<T>(obj: T): DeepUndefinedToNull<T> {
    return IsoObjectUtils.replace(obj, null, undefined);
  }

  static fullOuterJoin<K extends string, V>(
    left: Record<K, V | undefined>,
    right: Record<K, V | undefined>,
    reduce: (left: V, right: V) => V
  ): Record<K, V> {
    type VOpt = V | undefined;
    type Entry = [K, V];
    const isDefined = (entry: [string, VOpt]): entry is Entry => entry[1] !== undefined;
    const getEntries = (record: Record<K, VOpt>): Entry[] => Object.entries<VOpt>(record).filter(isDefined);
    const entries: Entry[] = [...getEntries(left), ...getEntries(right)];
    const entriesByKey: Record<K, Entry[]> = groupBy<Entry, K>(([k, _]) => k, entries);

    return mapObjIndexed(x => {
      if (x.length === 1) {
        return x[0][1];
      }
      if (x.length === 2) {
        return reduce(x[0][1], x[1][1]);
      }
      throw new Error(`Unexpected error: full outer join produced ${x.length} values, but expected 1 or 2.`);
    }, entriesByKey);
  }

  static unbox<T>(boxedDto: T): DeepUndefinedToNull<T> {
    return IsoObjectUtils.undefinedToNull(boxedDto);
  }

  static undefinedToNull<T>(obj: T): DeepUndefinedToNull<T> {
    return IsoObjectUtils.replace(obj, undefined, null);
  }

  private static replace<T, A, B>(obj: T, from: A, to: B): DeepUndefinedToNull<T> {
    if ((obj as any) === from) {
      return to as any;
    }

    // must go before 'is object' check, as arrays are objects.
    if (Array.isArray(obj)) {
      return (obj as any[]).map(x => IsoObjectUtils.replace(x, from, to)) as any;
    }

    if (typeof obj === "object") {
      if (obj === null) {
        return obj as any; // Null is treated as an object.
      }
      return Object.fromEntries(
        Object.entries(obj).map(([k, v]): [string, unknown] => [k, IsoObjectUtils.replace(v, from, to)])
      ) as any;
    }

    return obj as any;
  }
}
