import {
  FolderPath,
  FolderName,
  FileOrFolderPath,
  maybe,
  FilePath,
  FileName,
  RelativeOrAbsoluteFilePath,
  RelativeFilePath,
  RelativeOrAbsoluteFolderPath
} from "@upload-io/upload-shared-iso/common/model/Primitives";

/**
 * Extremely easy to make mistakes here, so be careful!
 *
 * (Fortunately we have unit tests for this class.)
 *
 * DO NOT USE Node's 'Path' object to perform the logic: these paths are S3-esque, so may contain "..", which may be
 * accidentally resolved if we use path libraries to do the parsing.
 */
export class FolderPathUtils {
  private static readonly allSlashesRegex = /^\/*$/;

  static splitExtension(fileName: FileName): [string, string] {
    const extensionSizeLimit = 4;
    const [name, ...parts] = fileName.split(".");
    let extension = "";
    let nameSuffix = "";
    let addToExtension = true;
    let part: string | undefined;
    while ((part = parts.pop()) !== undefined) {
      if (addToExtension && part.length <= extensionSizeLimit) {
        extension = `.${part}${extension}`;
      } else {
        addToExtension = false;
        nameSuffix = `.${part}${nameSuffix}`;
      }
    }
    return [`${name}${nameSuffix}`, extension];
  }

  static combineRelativeMaybe(
    parentFolder: FolderPath | undefined,
    childPath: RelativeOrAbsoluteFilePath | RelativeOrAbsoluteFolderPath
  ): FileOrFolderPath {
    if (FolderPathUtils.isAbsolute(childPath)) {
      return FileOrFolderPath(childPath);
    }
    return FolderPathUtils.combine(parentFolder, RelativeFilePath(childPath));
  }

  static isAbsolute(path: RelativeOrAbsoluteFilePath | RelativeOrAbsoluteFolderPath): boolean {
    return path.startsWith("/");
  }

  static isRelative(path: RelativeOrAbsoluteFilePath | RelativeOrAbsoluteFolderPath): boolean {
    return !FolderPathUtils.isAbsolute(path);
  }

  static combine(
    parentFolder: FolderPath | undefined,
    childName: RelativeFilePath | FolderName | FileName
  ): FileOrFolderPath {
    const addSeparator =
      parentFolder === undefined || !(FolderPathUtils.allSlashes(parentFolder) && childName.length > 0);
    return FileOrFolderPath(`${parentFolder ?? ""}${addSeparator ? "/" : ""}${childName}`);
  }

  static getFileName(filePath: FilePath): FileName {
    return FileName(FolderPathUtils.getName(filePath));
  }

  static getFolderName(folderPath: FolderPath): FolderName {
    return FolderName(FolderPathUtils.getName(folderPath));
  }

  static getName(folderPath: FolderPath | FilePath): FolderName | FileName {
    return FolderName(folderPath.substring(folderPath.lastIndexOf("/") + 1));
  }

  static getParentFolder(folderPath: FilePath | FolderPath | FileOrFolderPath): FolderPath | undefined {
    if (folderPath === "/") {
      return undefined;
    }

    const lastSlash = folderPath.lastIndexOf("/");
    const isTopLevelFolder = lastSlash === 0;
    const isUnnamedFolder = lastSlash === folderPath.length - 1;
    const unpaddedResult = isTopLevelFolder ? "/" : folderPath.substring(0, lastSlash);
    const consumeDash = isUnnamedFolder || isTopLevelFolder || !FolderPathUtils.allSlashes(unpaddedResult);
    const result = consumeDash ? unpaddedResult : `/${unpaddedResult}`;

    return FolderPath(result);
  }

  static getAncestors(folderPath: FilePath | FolderPath | FileOrFolderPath): FolderPath[] {
    if (folderPath === "/") {
      return [];
    }

    const result = [];
    let last: FolderPath | undefined;
    while ((last = FolderPathUtils.getParentFolder(maybe(FileOrFolderPath, last) ?? folderPath)) !== undefined) {
      result.push(last);
    }

    return result.reverse();
  }

  static allSlashes(value: string): boolean {
    return FolderPathUtils.allSlashesRegex.test(value);
  }

  /**
   * Gets the prefix that all descendants of this folder will have, that all non-descendants don't have.
   * Useful for translating folder paths to S3 prefixes.
   * Always ends with "/".
   */
  static getDescendantPrefix(parent: FolderPath): FolderPath {
    return FolderPathUtils.allSlashes(parent) ? parent : FolderPath(FolderPathUtils.combine(parent, FileName("")));
  }
}
