import { JSONSchema7 } from "json-schema";
import { StrictOmit } from "ts-essentials";
import { JsonPrimitive, NonEmptyArray } from "@upload-io/upload-shared-iso";

export type NormalizedJsonSchema = StrictOmit<JSONSchema7, "anyOf" | "allOf" | "oneOf" | "$ref">;

export interface ObjectFieldPossibleType {
  /**
   * If 'children' is non-empty, then we're implied to be inside an array or an object. Each 'child' represents a
   * field that has actually been set on the current array/object, so every element of 'possibleTypes' will have the
   * same number of children with the same names, however their own 'possibleTypes' may be different.
   */
  children: FieldWithMetadata[];

  /**
   * When the field represents an object property, this is the schema taken from the 'properties["foo"]'.
   * When the field represents an array item, this is the schema taken from the 'items' field.
   */
  schema: NormalizedJsonSchema;
}

/**
 * When TName is string: this type represents either an object's field, or an array's element.
 * When TName is undefined: this type represents a root object, array, or scalar.
 *
 * This type represents an object that is potentially partially populated, and contains all the possible types the
 * object could satisfy, should it be completed. The data structure uses a tree hierarchy, since adding values to new
 * fields higher in the object could invalidate entire subtrees, as the type is narrowed and potential candidates from
 * the original type union are elided.
 */
export interface FieldWithMetadata<TName = string, TTypes = NonEmptyArray<ObjectFieldPossibleType>> {
  /**
   * Name of a field on the object that has explicitly had a value set into it.
   * "0"/"1"/fieldX"/"fieldY"/etc.
   */
  fieldName: TName;

  /**
   * If value is an array or object, then the value of that array/object is expressed via 'possibleTypes[*].children'
   */
  fieldValue:
    | { source: "query" | "autofill"; type: "scalar"; value: JsonPrimitive }
    | { type: "array" }
    | { type: "object" };

  /**
   * The possible types this field may be, and for each of those types, each of its possible children and their possible types, recursively.
   */
  possibleTypes: TTypes;
}

export type ObjectWithMetadata = FieldWithMetadata<undefined>;
