import {
  FieldWithMetadata,
  ObjectFieldPossibleType
} from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/model/ObjectWithMetadata";
import { JSONSchema7 } from "json-schema";
import { JsonPrimitive, NonEmpty, optional } from "@upload-io/upload-shared-iso";
import { JsonSchemaUtils } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/utils/JsonSchemaUtils";
import { Incomplete, IsComplete } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/model/IsComplete";
import { sortBy } from "ramda";
import { QueryParamDefinitionFactory } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/utils/QueryParamDefinitionFactory";

export class ObjectAutoCompleter {
  /**
   * Fills in default and constant fields, recursively.
   *
   * Algorithm favours creating an entire object tree that is 100% complete (when determining which 'possibleTypes' to
   * use).
   */
  autoComplete<T>(object: FieldWithMetadata<T>): FieldWithMetadata<T> {
    return this.tryAutoComplete(object).value;
  }

  private getAutoFillValue(propertySchema: JSONSchema7): JsonPrimitive | undefined {
    return (
      propertySchema.default ?? ((propertySchema.enum?.length === 1 ? propertySchema.enum?.[0] : undefined) as any)
    );
  }

  private tryAutoComplete<T>(object: FieldWithMetadata<T>): IsComplete<FieldWithMetadata<T>> {
    if (object.fieldValue.type === "scalar") {
      return { value: object, isComplete: true };
    }

    const newPossibleTypes = object.possibleTypes.map(x => this.tryAutoCompletePossibleTypeNonScalar(x));
    const completePossibleTypes = newPossibleTypes.filter(x => x.isComplete);
    const isOuterComplete = completePossibleTypes.length > 0;
    const completePossibleTypesSorted = this.sortPossibleTypes(
      (object.fieldName as any) ?? "",
      completePossibleTypes,
      x => x.schema
    );
    return {
      isComplete: isOuterComplete,
      value: {
        ...object,
        // We only need to return one possibly type after auto completion.
        possibleTypes: [(isOuterComplete ? completePossibleTypesSorted : newPossibleTypes)[0].value]
      }
    };
  }

  private tryAutoCompletePossibleTypeNonScalar(
    possibleType: ObjectFieldPossibleType
  ): PossibleTypeAutoCompletionResult {
    return possibleType.schema.type === "array"
      ? this.tryAutoCompletePossibleTypeOnArray(possibleType)
      : this.tryAutoCompletePossibleTypeOnObject(possibleType);
  }

  private tryAutoCompletePossibleTypeOnArray(possibleType: ObjectFieldPossibleType): PossibleTypeAutoCompletionResult {
    const autoCompletedChildren = possibleType.children.map(x => this.tryAutoComplete(x));
    const isComplete = autoCompletedChildren.every(x => x.isComplete);
    return {
      isComplete,
      value: {
        ...possibleType,
        children: autoCompletedChildren.map(x => x.value)
      }
    };
  }

  private tryAutoCompletePossibleTypeOnObject(possibleType: ObjectFieldPossibleType): PossibleTypeAutoCompletionResult {
    const allProperties = possibleType.schema.properties ?? {};
    const existingProperties = possibleType.children;
    const definedPropertyNameSet = new Set(existingProperties.map(x => x.fieldName));
    const requiredPropertyNames = possibleType.schema.required ?? [];
    const requiredPropertyNamesSet = new Set(requiredPropertyNames);
    const remainingRequiredPropertyNames = requiredPropertyNames.filter(x => !definedPropertyNameSet.has(x));
    const remainingOptionalPropertiesWithDefaultValues = Object.entries(allProperties).flatMap(
      ([propertyName, propertySchema]) => {
        const isRemainingProperty = !definedPropertyNameSet.has(propertyName);
        const isOptionalProperty = !requiredPropertyNamesSet.has(propertyName);
        const isRemainingOptionalProperty = isRemainingProperty && isOptionalProperty;
        return isRemainingOptionalProperty && this.getAutoFillValue(propertySchema as JSONSchema7) !== undefined
          ? [propertyName]
          : [];
      }
    );

    const propertyNamesToAdd = [...remainingRequiredPropertyNames, ...remainingOptionalPropertiesWithDefaultValues];

    const newPropertiesMaybe = propertyNamesToAdd.map((propertyName):
      | IsComplete<FieldWithMetadata>
      | Incomplete<undefined> => {
      const propertyPossibleTypes = NonEmpty(
        JsonSchemaUtils.normalizePolymorphisms(allProperties[propertyName] as JSONSchema7).map(
          (type): ObjectFieldPossibleType => ({ schema: type, children: [] })
        )
      );

      const autoCompletedPropertyPossibleTypes = propertyPossibleTypes.map((propertyPossibleType):
        | IsComplete<FieldWithMetadata<string, ObjectFieldPossibleType>>
        | Incomplete<undefined> => {
        // We don't know the type yet, because of 'possibleTypes' has that information... it could be an array, object, or
        // other at this point! We need to do an 'isComplete' thing...
        if (propertyPossibleType.schema.type === "object" || propertyPossibleType.schema.type === "array") {
          const propertyAutoCompleted = this.tryAutoCompletePossibleTypeNonScalar(propertyPossibleType);
          return {
            isComplete: propertyAutoCompleted.isComplete,
            value: {
              fieldName: propertyName,
              fieldValue: { type: propertyPossibleType.schema.type },
              possibleTypes: propertyAutoCompleted.value
            }
          };
        }

        // Handle scalars...
        const autoFillValue = this.getAutoFillValue(propertyPossibleType.schema);
        if (autoFillValue === undefined) {
          return {
            isComplete: false,
            value: undefined
          };
        }
        return {
          isComplete: true,
          value: {
            fieldName: propertyName,
            fieldValue: { type: "scalar", value: autoFillValue, source: "autofill" },
            possibleTypes: propertyPossibleType
          }
        };
      });

      const propertyPossibleTypesSorted = this.sortPossibleTypes(
        propertyName,
        autoCompletedPropertyPossibleTypes,
        x => x?.possibleTypes?.schema
      );

      const completePropertyPossibleTypes = propertyPossibleTypesSorted.filter(x => x.isComplete);
      const value = (completePropertyPossibleTypes[0] ?? propertyPossibleTypesSorted[0]).value;
      if (value === undefined) {
        return {
          isComplete: false,
          value: undefined
        };
      }

      return {
        isComplete: completePropertyPossibleTypes.length > 0,
        value: {
          ...value,
          possibleTypes: [value.possibleTypes]
        }
      };
    });

    const newPropertiesAutoFilled = newPropertiesMaybe.flatMap(x => (x.value === undefined ? [] : [x]));
    const existingPropertiesAutoFilled = existingProperties.map(x => this.tryAutoComplete(x));
    const newPropertiesCompleted = newPropertiesMaybe.every(x => x.isComplete); // Must use 'newPropertiesMaybe' to take the 'isComplete: false' of undefined/un-addable properties into account.
    const existingPropertiesCompleted = existingPropertiesAutoFilled.every(x => x.isComplete);
    const autoCompletedChildren = [...existingPropertiesAutoFilled, ...newPropertiesAutoFilled];
    const isChildrenComplete = newPropertiesCompleted && existingPropertiesCompleted;

    return {
      isComplete: isChildrenComplete,
      value: {
        schema: possibleType.schema,
        children: autoCompletedChildren.map(x => x.value)
      }
    };
  }

  /**
   * Sorts all viable options to make the autocomplete behaviour deterministic. We do this by locating a field within
   * each type that has a 'Query Param Precedence: X' meta comment, and then sorting by X. If we can't find the field,
   * then we issue a warning, since this will result in a selection that could change if we add new types to the schema
   * in the future.
   */
  private sortPossibleTypes<T>(
    propertyName: string,
    types: Array<IsComplete<T>>,
    getSchema: (item: T) => JSONSchema7 | undefined
  ): Array<IsComplete<T>> {
    const hasMultipleViableObjectOptions =
      types.filter(x => x.isComplete && getSchema(x.value)?.type === "object").length > 1;
    if (!hasMultipleViableObjectOptions) {
      // Prevents issuing warnings when we're just dealing with a single type that doesn't have a 'Precedence' comment, etc.
      return types;
    }

    let hasUnknownPrecedence = false;

    const result = sortBy(x => {
      if (x.value === undefined || getSchema(x.value)?.type !== "object") {
        return Number.MAX_VALUE;
      }
      const propertySchemas = Object.values(getSchema(x.value)?.properties ?? {}) as JSONSchema7[];
      if (propertySchemas.length === 0) {
        return Number.MAX_VALUE;
      }

      const precedence = this.getQueryParamPrecedence(propertySchemas);
      if (precedence === undefined) {
        hasUnknownPrecedence = true;
        return Number.MAX_VALUE;
      }

      return precedence;
    }, types);

    if (hasUnknownPrecedence) {
      console.warn(
        `Property '${propertyName}' references a complex object that was autofilled. However, there were multiple possible types for this object, and at least one of these types didn't contain any properties that have a '${QueryParamDefinitionFactory.precedencePrefix}' comment, so the object that was chosen was non-deterministic.`
      );
    }

    return result;
  }

  private getQueryParamPrecedence(properties: JSONSchema7[]): number | undefined {
    return properties.flatMap((property): number[] =>
      optional(QueryParamDefinitionFactory.getQueryParamPrecedence(property))
    )[0];
  }
}

type PossibleTypeAutoCompletionResult = IsComplete<ObjectFieldPossibleType>;
