import { QueryParamDefinitionMap } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/model/QueryParamDefinitionMap";
import { JSONSchema7 } from "json-schema";
import { QueryParamDefinition } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/model/QueryParamDefinition";
import { groupBy, uniq, map, sortBy, mergeAll } from "ramda";
import { JsonSchemaUtils } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/utils/JsonSchemaUtils";

export class QueryParamDefinitionFactory {
  private static readonly namePrefix = "Query Param: ";
  private static readonly constantRewritePrefix = "Query Param Value: ";
  static readonly precedencePrefix = "Query Param Precedence: ";

  private static readonly precedence: Record<QueryParamDefinition["type"], number> = {
    null: 0,
    number: 1,
    boolean: 2,
    string: 3
  };

  static createDefinitions(schema: JSONSchema7): QueryParamDefinitionMap {
    const queryParams = QueryParamDefinitionFactory.conflateQueryParams(
      QueryParamDefinitionFactory.createDefinitionEntries(schema, [])
    );

    const queryParamsByName = groupBy(x => x.queryParamName, queryParams);

    const badOverloadedQueryParams = Object.values(queryParamsByName).flatMap(overloads => {
      const overloadsOverlap = overloads.length !== uniq(overloads.map(x => x.type)).length;
      return overloadsOverlap ? [overloads] : [];
    });

    if (badOverloadedQueryParams.length > 0) {
      throw new Error(
        `Detected duplicate query param names: ${badOverloadedQueryParams
          .map(x => `${x[0].queryParamName} (${x.map(y => `${y.type}:${y.jsonPath.join(".")}`).join(", ")})`)
          .join(
            ", "
          )}.\nEach of these query params references multiple different JSON paths, but use the same type for some of the paths, so cannot be distinguished.`
      );
    }

    return map(x => QueryParamDefinitionFactory.sortQueryParamsByTypePrecedence(x), queryParamsByName);
  }

  static getQueryParamName(schema: JSONSchema7): string | undefined {
    return QueryParamDefinitionFactory.getQuerySchemaMetadata(schema, QueryParamDefinitionFactory.namePrefix);
  }

  static getQueryParamConstantRewrite(schema: JSONSchema7): string | undefined {
    return QueryParamDefinitionFactory.getQuerySchemaMetadata(
      schema,
      QueryParamDefinitionFactory.constantRewritePrefix
    );
  }

  static getQueryParamPrecedence(schema: JSONSchema7): number | undefined {
    const value = QueryParamDefinitionFactory.getQuerySchemaMetadata(
      schema,
      QueryParamDefinitionFactory.precedencePrefix
    );
    if (value === undefined) {
      return undefined;
    }

    const valueNum = parseInt(value);
    if (isNaN(valueNum)) {
      return undefined;
    }

    return valueNum;
  }

  private static getQuerySchemaMetadata(schema: JSONSchema7, commentPrefix: string): string | undefined {
    const commentLine = (schema.description ?? "")
      .split("\n")
      .map(x => x.trim())
      .find(x => x.startsWith(commentPrefix));

    if (commentLine === undefined) {
      return undefined;
    }

    return commentLine.substring(commentPrefix.length);
  }

  /**
   * Combine query params that have the same name, same type, and same JSON path. This can occur on polymorphic types,
   * e.g. the "type" discriminator on 5 different types within a union will result in 5 identical query param definitions.
   */
  private static conflateQueryParams(params: QueryParamDefinition[]): QueryParamDefinition[] {
    return Object.values(groupBy(x => JSON.stringify([x.queryParamName, x.type, x.jsonPath]), params)).map(
      (group): QueryParamDefinition => ({
        ...group[0],
        rewrites: mergeAll(group.map(x => x.rewrites))
      })
    );
  }

  /**
   * If a query param references multiple property/type pairs, then sort by the order in which we want to attempt
   * deserialization in. Since querystring values arrive as strings, if a query param supports numbers _or_ strings,
   * we need to ensure deserialization is first attempted with the number type, and then fall back onto string. This
   * method essentially sorts the query param definitions by the order in which deserialization should be attempted in,
   * from most narrow to broad.
   */
  private static sortQueryParamsByTypePrecedence(params: QueryParamDefinition[]): QueryParamDefinition[] {
    return sortBy(x => QueryParamDefinitionFactory.precedence[x.type], params);
  }

  private static createDefinitionEntries(schemaRaw: JSONSchema7, jsonPath: string[]): QueryParamDefinition[] {
    const schemas = JsonSchemaUtils.normalizePolymorphisms(schemaRaw);
    if (schemas.length > 1) {
      return schemas.flatMap(x => QueryParamDefinitionFactory.createDefinitionEntries(x, jsonPath));
    }
    const schema = schemas[0];
    if (schema.items !== undefined) {
      return QueryParamDefinitionFactory.createDefinitionEntries(schema.items as JSONSchema7, [...jsonPath, "*"]);
    }
    if (schema.properties !== undefined) {
      return Object.entries(schema.properties).flatMap(([propertyName, propertySchema]) =>
        QueryParamDefinitionFactory.createDefinitionEntries(propertySchema as JSONSchema7, [...jsonPath, propertyName])
      );
    }

    const queryParamName = QueryParamDefinitionFactory.getQueryParamName(schema);

    if (queryParamName === undefined) {
      return [];
    }

    if (
      !(
        schema.type === "string" ||
        schema.type === "boolean" ||
        schema.type === "number" ||
        schema.type === "integer" ||
        schema.type === "null"
      )
    ) {
      return [];
    }

    const rewrites: Record<string, string | undefined> = {};
    const constantRewriteFrom = QueryParamDefinitionFactory.getQueryParamConstantRewrite(schema);
    const constantRewriteTo = schema.enum !== undefined && schema.enum.length === 1 ? schema.enum[0] : undefined;
    if (
      constantRewriteFrom !== undefined &&
      constantRewriteTo !== undefined &&
      (typeof constantRewriteTo === "string" ||
        typeof constantRewriteTo === "number" ||
        typeof constantRewriteTo === "boolean")
    ) {
      rewrites[constantRewriteFrom] = constantRewriteTo.toString();
    }

    return [{ queryParamName, type: schema.type === "integer" ? "number" : schema.type, jsonPath, rewrites }];
  }
}
