import { JSONSchema7 } from "json-schema";
import {
  OrderSensitiveQueryParams,
  QueryParam
} from "@upload-io/upload-shared-iso/common/model/OrderSensitiveQueryParams";
import { assertUnreachable, isDefined, JsonPrimitive, NonEmpty, NonEmptyMaybe } from "@upload-io/upload-shared-iso";
import { QueryParamDefinitionMap } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/model/QueryParamDefinitionMap";
import { ObjectAutoCompleter } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/utils/ObjectAutoCompleter";
import {
  FieldWithMetadata,
  ObjectFieldPossibleType,
  ObjectWithMetadata
} from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/model/ObjectWithMetadata";
import { QueryParamDefinition } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/model/QueryParamDefinition";
import { ObjectRasterizer } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/utils/ObjectRasterizer";
import { JsonSchemaUtils } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/utils/JsonSchemaUtils";
import { QueryStringToJsonResult } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/model/QueryStringToJsonResult";

export class QueryStringToJsonFunction {
  private readonly objectAutoCompleter: ObjectAutoCompleter;
  private readonly objectRasterizer: ObjectRasterizer;

  constructor(private readonly schema: JSONSchema7, private readonly paramDefinitions: QueryParamDefinitionMap) {
    this.objectAutoCompleter = new ObjectAutoCompleter();
    this.objectRasterizer = new ObjectRasterizer();
  }

  /**
   * IMPORTANT: returned object may be invalid, due to being incomplete (as there may not have been enough data given in
   * the query params) or may be invalid due to extended validation rules like regex patterns and min/max/minItems/maxItems
   * rules not being checked. Essentially, we only check basic typing and object structure, up to the point it effects
   * type union resolution, but nothing beyond that.
   */
  toJson(queryString: OrderSensitiveQueryParams): QueryStringToJsonResult {
    const invalidParamCombinations: string[] = [];
    const invalidParamTypes: string[] = [];
    const unknownParamNames: string[] = [];
    const finalResult = queryString.reduce<ObjectWithMetadata>(
      (result, param): ObjectWithMetadata => {
        const paramResult = this.setQueryParam(result, param);
        switch (paramResult.type) {
          case "success":
            return paramResult.result;
          case "invalidParamType":
            invalidParamTypes.push(param.name);
            return result;
          case "invalidParamCombination":
            invalidParamCombinations.push(param.name);
            return result;
          case "unknownParamName":
            unknownParamNames.push(param.name);
            return result;
          default:
            assertUnreachable(paramResult);
        }
      },
      {
        fieldName: undefined,
        fieldValue: { type: "object" },
        possibleTypes: NonEmpty(
          JsonSchemaUtils.normalizePolymorphisms(this.schema).map(
            (type): ObjectFieldPossibleType => ({ schema: type, children: [] })
          )
        )
      }
    );
    const completedObject = this.objectAutoCompleter.autoComplete(finalResult);
    const rasterizedObject = this.objectRasterizer.rasterize(completedObject);
    return { result: rasterizedObject, unknownParamNames, invalidParamCombinations, invalidParamTypes };
  }

  private setQueryParam(
    root: ObjectWithMetadata,
    param: QueryParam
  ):
    | { result: ObjectWithMetadata; type: "success" }
    | { type: "invalidParamType" }
    | { type: "invalidParamCombination" }
    | { type: "unknownParamName" } {
    const paramDefs = this.paramDefinitions[param.name];
    if (paramDefs === undefined) {
      return { type: "unknownParamName" };
    }

    // Handle query param overloading by parsing the query param value, and using the first def that can hold the param's value (if any).
    const paramDefAndValue = paramDefs.flatMap(
      (paramDef): Array<{ paramDef: QueryParamDefinition; paramValue: JsonPrimitive }> => {
        const paramValue = this.tryParseParamValue(param, paramDef);
        return paramValue === undefined ? [] : [{ paramValue, paramDef }];
      }
    )[0];

    if (paramDefAndValue === undefined) {
      return { type: "invalidParamType" };
    }

    const { paramValue, paramDef } = paramDefAndValue;

    const updatedRoot = this.trySet(paramValue, paramDef.jsonPath, root);
    if (updatedRoot === undefined) {
      // Returned when an invalid combination of query parameters are given.
      return { type: "invalidParamCombination" };
    }

    return { type: "success", result: updatedRoot };
  }

  /**
   * Returns the updated 'object' or 'undefined' if it's impossible to set the given field into this object.
   */
  private trySet<T>(
    queryParamValue: JsonPrimitive,
    fieldPath: string[],
    object: FieldWithMetadata<T>
  ): FieldWithMetadata<T> | undefined {
    if (fieldPath.length === 0) {
      // Process as scalar.
      const passesOwnTypeChecks = (possibleType: ObjectFieldPossibleType): boolean =>
        JsonSchemaUtils.passesBasicTypeCheck(possibleType.schema, queryParamValue);

      const newPossibleTypes = object.possibleTypes.filter(passesOwnTypeChecks);
      const newPossibleTypesMaybe = NonEmptyMaybe(newPossibleTypes);

      return newPossibleTypesMaybe === undefined
        ? undefined
        : {
            ...object,
            possibleTypes: newPossibleTypesMaybe
          };
    }

    // Process as complex object or array.
    const [childFieldNameRaw, ...tailFieldPath] = fieldPath;
    const isChildFieldScalar = fieldPath.length === 1; // We're setting a value to a field on this object/array, so it's a scalar.
    const isChildFieldArray = fieldPath.length > 1 && fieldPath[1] === "*";
    const isChildFieldArrayElement = childFieldNameRaw === "*";

    // Checks if new childField combined with the existing children satisfy the possibleType's schema, but only performs
    // local checks, meaning "is the 'type' correct?" and "can these field names coexist?". Checking descendent types
    // and bubbling up incompatibles caused by descendents we're adding is done on the recursive flatMap phase, not
    // this filter phase.
    const passesOwnTypeChecks = (childFieldName: string) => (possibleType: ObjectFieldPossibleType): boolean => {
      const ownSchema = possibleType.schema;
      const ownType = ownSchema.type;
      if (ownType === "array") {
        return isChildFieldArrayElement;
      }
      if (ownType === "object") {
        const existingChildNames = possibleType.children.map(x => x.fieldName);
        const newChildNames = [...existingChildNames, childFieldName];
        const newChildNamesViable = newChildNames.every(x => ownSchema.properties?.[x] !== undefined);
        const childAlreadyExists = existingChildNames.includes(childFieldName);
        // If child with same name already exists AND I am a leaf, then bail, since we can't overwrite existing values, we must instead prompt new array elements to be created.
        const attemptToOverwriteScalar = childAlreadyExists && isChildFieldScalar;
        return !attemptToOverwriteScalar && newChildNamesViable;
      }

      // Happens when the existing field is a scalar, but we need an array or complex object, in which case it's a type mismatch, so does not pass type checks.
      return false;
    };

    // At this point, we've confirmed current object is a complex object or an array (so can accept the child field)
    // and that we're allows to create a child field with this name, or it already exists and we're allowed to traverse
    // into it.
    const tryAddChild = (childFieldName: string) => (
      possibleType: ObjectFieldPossibleType
    ): [] | [ObjectFieldPossibleType] => {
      const ownSchema = possibleType.schema;
      const ownType = ownSchema.type;
      const childAlreadyExists = possibleType.children.some(x => x.fieldName === childFieldName);

      const childFieldSchema = ownType === "array" ? ownSchema.items : ownSchema.properties?.[childFieldName];
      if (childFieldSchema === undefined) {
        throw new Error("childFieldSchema was undefined: unexpected due to earlier checks in this function.");
      }
      if (typeof childFieldSchema !== "object" || Array.isArray(childFieldSchema)) {
        throw new Error(
          `childFieldSchema was of unexpected type '${(typeof childFieldSchema !== "object").toString()}'.`
        );
      }

      const updatedChildren = childAlreadyExists
        ? possibleType.children.map(x =>
            x.fieldName === childFieldName ? this.trySet<string>(queryParamValue, tailFieldPath, x) : x
          )
        : [
            ...possibleType.children,
            this.trySet<string>(queryParamValue, tailFieldPath, {
              fieldName: childFieldName,
              fieldValue: isChildFieldScalar
                ? { type: "scalar", value: queryParamValue, source: "query" }
                : isChildFieldArray
                ? { type: "array" }
                : { type: "object" },
              possibleTypes: NonEmpty(
                JsonSchemaUtils.normalizePolymorphisms(childFieldSchema).map(
                  (type): ObjectFieldPossibleType => ({ schema: type, children: [] })
                )
              )
            })
          ];

      // If the child was inviable, it will have left an 'undefined' element in this collection. If so, this entire
      // subtree is not viable, so return empty array to elide it.
      const updatedChildrenDefined: FieldWithMetadata[] = updatedChildren.filter(isDefined);
      return updatedChildrenDefined.length !== updatedChildren.length
        ? []
        : [
            {
              schema: possibleType.schema,
              children: updatedChildrenDefined
            }
          ];
    };

    const childFieldName = isChildFieldArrayElement
      ? this.calculateNextArrayElementIndex(object, isChildFieldScalar, false).toString()
      : childFieldNameRaw;
    const newPossibleTypesBase = object.possibleTypes.filter(passesOwnTypeChecks(childFieldName));

    // In the context of the current type being an array, the below will attempt to add the child field to the last
    // element that exists in the array (or create one if none exists), see 'calculateNextArrayElementIndex' above.
    const newPossibleTypesFirstAttempt = newPossibleTypesBase.flatMap(tryAddChild(childFieldName));

    // In the context of the current type being an array, if the above attempt at adding the child field failed, it may
    // have been due to the existing array element (which we were adding the field to) already being refined to a subset
    // of a potential type union that doesn't support the new field, in which case, we try again but this time adding a
    // new array element to add the field to.
    const failedToSetChildField = newPossibleTypesFirstAttempt.length === 0 && newPossibleTypesBase.length > 0;
    const failedToSetArrayElement = failedToSetChildField && isChildFieldArrayElement;
    const newPossibleTypes = failedToSetArrayElement
      ? newPossibleTypesBase.flatMap(
          tryAddChild(this.calculateNextArrayElementIndex(object, isChildFieldScalar, true).toString())
        )
      : newPossibleTypesFirstAttempt;

    const newPossibleTypesMaybe = NonEmptyMaybe(newPossibleTypes);
    return newPossibleTypesMaybe === undefined
      ? undefined
      : {
          ...object,
          possibleTypes: newPossibleTypesMaybe
        };
  }

  /**
   * If the array is empty, or we're adding a scalar element, or we forcefully want a new array element, then create
   * a new element, else return the last element, so it can be updated in-place (which we always want to attempt to do
   * first if it's a complex object, i.e. attempt augment the last element, and on failure, use forceAppend to add new).
   */
  private calculateNextArrayElementIndex(
    array: FieldWithMetadata<unknown>,
    isScalar: boolean,
    forceAppend: boolean
  ): number {
    return Math.max(0, array.possibleTypes[0].children.length + (isScalar || forceAppend ? 0 : -1));
  }

  private tryParseParamValue(param: QueryParam, paramDef: QueryParamDefinition): JsonPrimitive | undefined {
    const value = paramDef.rewrites[param.value] ?? param.value;

    switch (paramDef.type) {
      case "null":
        return value === "null" ? null : undefined;

      case "number": {
        const numberMaybe = parseInt(value);
        return !isNaN(numberMaybe) ? numberMaybe : undefined;
      }

      case "boolean": {
        const lowerTrimmed = value.toLowerCase().trim();
        return lowerTrimmed === true.toString() ? true : lowerTrimmed === false.toString() ? false : undefined;
      }

      case "string": {
        return value;
      }

      default:
        assertUnreachable(paramDef.type);
    }
  }
}
