import { groupBy, mapObjIndexed, sum } from "ramda";
import { StorageStatisticsSegmentDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/StorageStatisticsSegmentDto";
import { UploadStatisticsDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsDto";
import { BigSizeInBytes, EpochMillis, FileTagName } from "@upload-io/upload-shared-iso/common/model/Primitives";
import { Milliseconds, SizeInBytes } from "@upload-io/upload-shared-iso-lite";
import { NonEmptyArray } from "@upload-io/upload-shared-iso/common/model/NonEmptyArray";
import { BasicFileEvent } from "@upload-io/upload-shared-iso/statistics/storage/BasicFileEvent";
import { IsoTimeUtils } from "@upload-io/upload-shared-iso/common/IsoTimeUtils";
import { BoxedArithmetic } from "@upload-io/upload-shared-iso/common/BoxedArithmetic";
import { BytesByEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/BytesByEpochDto";
import { DailyMonthlyBytesDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DailyMonthlyBytesDto";
import { StorageStatisticsByTimeDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/StorageStatisticsByTimeDto";
import { InvocationEvent } from "@upload-io/upload-shared-iso/statistics/storage/InvocationEvent";
import { DurationByEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DurationByEpochDto";
import { DailyMonthlyDurationsDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DailyMonthlyDurationsDto";
import { DailyMonthlyUploadsDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DailyMonthlyUploadsDto";
import { UploadsByEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadsByEpochDto";
import { UploadsAtEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadsAtEpochDto";

const daysToRecord = 60; // UI requires at least 30 for 'last 30 days' widget. We capture more for good measure.
const monthsToRecord = 24; // UI requires at least 12 for 'monthly bandwidth' widget. We capture more for good measure.

export class StatisticsCalculator {
  private static readonly dashboardUploadTag = FileTagName("manual_upload"); // Sync with: upload-website

  static calculateStatistics(
    newEvents: NonEmptyArray<BasicFileEvent>,
    sizeMultiplier: 1 | -1 | 0,
    countMultiplier: 1 | -1 | 0
  ): UploadStatisticsDto {
    return {
      all: this.calculateStatisticsSegment(newEvents, sizeMultiplier, countMultiplier),
      byTime: this.calculateStorageStatisticsByTime(newEvents, sizeMultiplier, countMultiplier)
    };
  }

  static calculateMonthlyDailyBytes(events: BasicFileEvent[]): DailyMonthlyBytesDto {
    const now = EpochMillis(Date.now());
    return {
      byDay: this.calculateBytesPerEpochUnit(IsoTimeUtils.toEpochDay, daysToRecord, events, now),
      byMonth: this.calculateBytesPerEpochUnit(IsoTimeUtils.toEpochMonth, monthsToRecord, events, now)
    };
  }

  static calculateMonthlyDailyUploads(events: BasicFileEvent[]): DailyMonthlyUploadsDto {
    const now = EpochMillis(Date.now());
    return {
      byDay: this.calculateUploadsPerEpochUnit(IsoTimeUtils.toEpochDay, daysToRecord, events, now),
      byMonth: this.calculateUploadsPerEpochUnit(IsoTimeUtils.toEpochMonth, monthsToRecord, events, now)
    };
  }

  static calculateMonthlyDailyDurations(events: InvocationEvent[]): DailyMonthlyDurationsDto {
    const now = EpochMillis(Date.now());
    return {
      byDay: this.calculateDurationsPerEpochUnit(IsoTimeUtils.toEpochDay, daysToRecord, events, now),
      byMonth: this.calculateDurationsPerEpochUnit(IsoTimeUtils.toEpochMonth, monthsToRecord, events, now)
    };
  }

  static calculateStatisticsSegment(
    newEvents: NonEmptyArray<BasicFileEvent>,
    sizeMultiplier: 1 | -1 | 0,
    countMultiplier: 1 | -1 | 0
  ): StorageStatisticsSegmentDto {
    const timestamps = newEvents.map(x => x.timeOfEvent);
    const eventSizeSum = sum(newEvents.map(x => x.size));
    const eventCount = newEvents.length;
    return {
      firstFileCreated: sizeMultiplier === -1 ? undefined : EpochMillis(Math.min(...timestamps)),
      lastFileCreated: sizeMultiplier === -1 ? undefined : EpochMillis(Math.max(...timestamps)),
      totalFileCount: countMultiplier * eventCount,
      totalFileCountMonotonic: Math.max(0, countMultiplier) * eventCount,
      totalFileSize: SizeInBytes(sizeMultiplier * eventSizeSum), // Can go negative (in a shard)
      totalFileSizeMonotonic: SizeInBytes(Math.max(0, sizeMultiplier) * eventSizeSum)
    };
  }

  static isFileFromDashboard(tags: FileTagName[]): boolean {
    return tags.some(x => x === this.dashboardUploadTag);
  }

  static calculateStorageStatisticsByTime(
    newEvents: NonEmptyArray<BasicFileEvent>,
    sizeMultiplier: 1 | -1 | 0,
    countMultiplier: 1 | -1 | 0
  ): StorageStatisticsByTimeDto {
    return {
      creates: sizeMultiplier === 1 ? this.calculateMonthlyDailyBytes(newEvents) : DailyMonthlyBytesDto.empty,
      createsExtended:
        countMultiplier === 1 ? this.calculateMonthlyDailyUploads(newEvents) : DailyMonthlyUploadsDto.empty,
      deletes: sizeMultiplier === -1 ? this.calculateMonthlyDailyBytes(newEvents) : DailyMonthlyBytesDto.empty
    };
  }

  private static calculateBytesPerEpochUnit(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    events: BasicFileEvent[],
    now: EpochMillis
  ): BytesByEpochDto {
    const totalSize = (events: BasicFileEvent[]): BigSizeInBytes =>
      BoxedArithmetic.bigSizeInBytes.sum(events.map(event => BigSizeInBytes(event.size.toString())));

    return StatisticsCalculator.reducePerEpochUnit(getEpochUnit, keepLastNUnits, events, now, totalSize);
  }

  private static calculateUploadsPerEpochUnit(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    events: BasicFileEvent[],
    now: EpochMillis
  ): UploadsByEpochDto {
    const totalCount = (events: BasicFileEvent[]): UploadsAtEpochDto => ({
      totalCount: events.length
    });

    return StatisticsCalculator.reducePerEpochUnit(getEpochUnit, keepLastNUnits, events, now, totalCount);
  }

  private static reducePerEpochUnit<T>(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    events: BasicFileEvent[],
    now: EpochMillis,
    reduce: (group: BasicFileEvent[]) => T
  ): {
    [epochValue: string]: T | undefined;
  } {
    const eventsByEpoch = groupBy<BasicFileEvent>(x => getEpochUnit(x.timeOfEvent).toString(), events);
    return IsoTimeUtils.trimEpochKeys(getEpochUnit, now, keepLastNUnits, mapObjIndexed(reduce, eventsByEpoch));
  }

  private static calculateDurationsPerEpochUnit(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    events: InvocationEvent[],
    now: EpochMillis
  ): DurationByEpochDto {
    const eventsByEpoch = groupBy<InvocationEvent>(x => getEpochUnit(x.timeOfEvent).toString(), events);
    const totalSize = (events: InvocationEvent[]): Milliseconds =>
      Milliseconds(sum(events.map(event => event.duration)));

    return IsoTimeUtils.trimEpochKeys(getEpochUnit, now, keepLastNUnits, mapObjIndexed(totalSize, eventsByEpoch));
  }
}
