/* eslint-disable no-template-curly-in-string */
import { JsonObjectUnboxed } from "@upload-io/upload-shared-iso";
import imagePluginSchema from "./ExamplePluginSchema.ImagePlugin.json";
import compressionPluginSchema from "./ExamplePluginSchema.CompressionPlugin.json";
import { OrderSensitiveQueryParams } from "@upload-io/upload-shared-iso/common/model/OrderSensitiveQueryParams";
import { QueryStringToJsonMapper } from "@upload-io/upload-shared-iso/plugins/qsToJsonMapper/QueryStringToJsonMapper";
import { JSONSchema7 } from "json-schema";

describe("QueryStringToJsonMapper", (): void => {
  test("empty querystring", async (): Promise<void> => {
    await assertTwoWayMap(imagePluginSchema as JSONSchema7, [], [], { pipeline: { steps: [] } });
  });

  test("simple querystring", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [{ name: "f", value: "webp" }],
      [{ name: "f", value: "webp" }],
      {
        pipeline: { steps: [], outputFormat: "webp" }
      }
    );
  });

  test("allow the same query param name to reference two different field paths, providing the query param types are mutually exclusive", async (): Promise<
    void
  > => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [{ name: "rotate", value: "90" }],
      [
        { name: "rotate", value: "90" },
        { name: "rotate-bg", value: "00000000" }
      ],
      {
        pipeline: {
          steps: [
            {
              type: "rotate",
              mode: {
                backgroundColor: "#00000000",
                degrees: 90,
                type: "manual"
              }
            }
          ]
        }
      }
    );
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [{ name: "rotate", value: "auto" }],
      [{ name: "rotate", value: "auto" }],
      {
        pipeline: {
          steps: [
            {
              type: "rotate",
              mode: {
                type: "automatic"
              }
            }
          ]
        }
      }
    );
  });

  // Since hashes can't appear in server-side URLs, we need to support hashes being omitted in the query strings, meaning we must re-add them here.
  test("automatically prepend hash prefixes to values that require them", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [{ name: "tint", value: "00cc00" }],
      [{ name: "tint", value: "00cc00" }],
      {
        pipeline: {
          steps: [
            {
              type: "tint",
              color: "#00cc00"
            }
          ]
        }
      }
    );
  });

  test("infer type discriminator & defaults from regular field", async (): Promise<void> => {
    await assertTwoWayMap(
      compressionPluginSchema as JSONSchema7,
      [{ name: "compression", value: "8" }],
      [
        {
          name: "compression",
          value: "8"
        },
        {
          name: "format",
          value: "zip"
        },
        { name: "mode", value: "archive" }
      ],
      {
        mode: "archive",
        settings: {
          format: "zip",
          compressionLevel: 8
        }
      }
    );
  });

  test("infer an entire nested field, where the nested field is a polymorphic type, and only one of the types has all fields defaulted", async (): Promise<
    void
  > => {
    await assertTwoWayMap(
      compressionPluginSchema as JSONSchema7,
      [{ name: "mode", value: "archive" }],
      [
        {
          name: "compression",
          value: "7"
        },
        {
          name: "format",
          value: "zip"
        },
        { name: "mode", value: "archive" }
      ],
      {
        mode: "archive",
        settings: {
          format: "zip",
          compressionLevel: 7
        }
      }
    );
  });

  test("infer most relevant type discriminator (test 1)", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [{ name: "w", value: "400" }],
      [
        {
          name: "fit",
          value: "width"
        },
        { name: "w", value: "400" }
      ],
      {
        pipeline: {
          steps: [
            {
              type: "resize",
              geometry: {
                size: {
                  type: "width",
                  width: 400
                }
              }
            }
          ]
        }
      }
    );
  });

  test("infer most relevant type discriminator (test 2)", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [{ name: "h", value: "300" }],
      [
        {
          name: "fit",
          value: "height"
        },
        { name: "h", value: "300" }
      ],
      {
        pipeline: {
          steps: [
            {
              type: "resize",
              geometry: {
                size: {
                  type: "xheight",
                  height: 300
                }
              }
            }
          ]
        }
      }
    );
  });

  test("infer most relevant type discriminator (test 3)", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [
        { name: "w", value: "400" },
        { name: "h", value: "300" }
      ],
      [
        {
          name: "fit",
          value: "max"
        },
        {
          name: "h",
          value: "300"
        },
        {
          name: "w",
          value: "400"
        }
      ],
      {
        pipeline: {
          steps: [
            {
              type: "resize",
              geometry: {
                size: {
                  type: "widthxheight",
                  width: 400,
                  height: 300
                }
              }
            }
          ]
        }
      }
    );
  });

  test("infer most relevant type discriminator (test 4)", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [
        { name: "w", value: "400" },
        { name: "h", value: "300" },
        { name: "fit", value: "stretch" }
      ],
      [
        { name: "fit", value: "stretch" },
        { name: "h", value: "300" },
        { name: "w", value: "400" }
      ],
      {
        pipeline: {
          steps: [
            {
              type: "resize",
              geometry: {
                size: {
                  type: "widthxheight!",
                  width: 400,
                  height: 300
                }
              }
            }
          ]
        }
      }
    );
  });

  test("respect order of querystring params (test 1)", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [
        { name: "w", value: "400" },
        { name: "h", value: "300" },
        { name: "fit", value: "stretch" },
        { name: "rotate", value: "90" }
      ],
      [
        { name: "fit", value: "stretch" },
        { name: "h", value: "300" },
        { name: "w", value: "400" },
        { name: "rotate", value: "90" },
        {
          name: "rotate-bg",
          value: "00000000"
        }
      ],
      {
        pipeline: {
          steps: [
            {
              type: "resize",
              geometry: {
                size: {
                  type: "widthxheight!",
                  width: 400,
                  height: 300
                }
              }
            },
            {
              type: "rotate",
              mode: {
                type: "manual",
                degrees: 90,
                backgroundColor: "#00000000"
              }
            }
          ]
        }
      }
    );
  });

  test("respect order of querystring params (test 2)", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [
        { name: "rotate", value: "90" },
        { name: "w", value: "400" },
        { name: "h", value: "300" },
        { name: "fit", value: "stretch" }
      ],
      [
        { name: "rotate", value: "90" },
        { name: "rotate-bg", value: "00000000" },
        { name: "fit", value: "stretch" },
        { name: "h", value: "300" },
        { name: "w", value: "400" }
      ],
      {
        pipeline: {
          steps: [
            {
              type: "rotate",
              mode: {
                type: "manual",
                degrees: 90,
                backgroundColor: "#00000000"
              }
            },
            {
              type: "resize",
              geometry: {
                size: {
                  type: "widthxheight!",
                  width: 400,
                  height: 300
                }
              }
            }
          ]
        }
      }
    );
  });

  test("respect order of querystring params (test 3)", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [
        { name: "w", value: "400" },
        { name: "rotate", value: "90" },
        { name: "h", value: "300" }
      ],
      [
        { name: "fit", value: "width" },
        { name: "w", value: "400" },
        { name: "rotate", value: "90" },
        { name: "rotate-bg", value: "00000000" },
        { name: "fit", value: "height" },
        { name: "h", value: "300" }
      ],
      {
        pipeline: {
          steps: [
            {
              type: "resize",
              geometry: {
                size: {
                  type: "width",
                  width: 400
                }
              }
            },
            {
              type: "rotate",
              mode: {
                type: "manual",
                degrees: 90,
                backgroundColor: "#00000000"
              }
            },
            {
              type: "resize",
              geometry: {
                size: {
                  type: "xheight",
                  height: 300
                }
              }
            }
          ]
        }
      }
    );
  });

  test("handle repetitions of same element type (test 1)", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [
        { name: "rotate", value: "90" },
        { name: "rotate", value: "180" },
        { name: "w", value: "400" },
        { name: "h", value: "300" },
        { name: "fit", value: "stretch" }
      ],
      [
        { name: "rotate", value: "90" },
        { name: "rotate-bg", value: "00000000" },
        { name: "rotate", value: "180" },
        { name: "rotate-bg", value: "00000000" },
        { name: "fit", value: "stretch" },
        { name: "h", value: "300" },
        { name: "w", value: "400" }
      ],
      {
        pipeline: {
          steps: [
            {
              type: "rotate",
              mode: {
                type: "manual",
                degrees: 90,
                backgroundColor: "#00000000"
              }
            },
            {
              type: "rotate",
              mode: {
                type: "manual",
                degrees: 180,
                backgroundColor: "#00000000"
              }
            },
            {
              type: "resize",
              geometry: {
                size: {
                  type: "widthxheight!",
                  width: 400,
                  height: 300
                }
              }
            }
          ]
        }
      }
    );
  });

  test("handle repetitions of same element type (test 2)", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [
        { name: "rotate-bg", value: "ccccccaa" },
        { name: "rotate", value: "90" },
        { name: "rotate", value: "180" },
        { name: "w", value: "400" },
        { name: "h", value: "300" },
        { name: "fit", value: "stretch" }
      ],
      [
        { name: "rotate", value: "90" },
        { name: "rotate-bg", value: "ccccccaa" },
        { name: "rotate", value: "180" },
        { name: "rotate-bg", value: "00000000" },
        { name: "fit", value: "stretch" },
        { name: "h", value: "300" },
        { name: "w", value: "400" }
      ],
      {
        pipeline: {
          steps: [
            {
              type: "rotate",
              mode: {
                type: "manual",
                degrees: 90,
                backgroundColor: "#ccccccaa"
              }
            },
            {
              type: "rotate",
              mode: {
                type: "manual",
                degrees: 180,
                backgroundColor: "#00000000"
              }
            },
            {
              type: "resize",
              geometry: {
                size: {
                  type: "widthxheight!",
                  width: 400,
                  height: 300
                }
              }
            }
          ]
        }
      }
    );
  });

  test("handle repetitions of same element type (test 3)", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [
        { name: "rotate", value: "90" },
        { name: "rotate", value: "auto" },
        { name: "w", value: "400" },
        { name: "h", value: "300" },
        { name: "fit", value: "stretch" }
      ],
      [
        { name: "rotate", value: "90" },
        { name: "rotate-bg", value: "00000000" },
        { name: "rotate", value: "auto" },
        { name: "fit", value: "stretch" },
        { name: "h", value: "300" },
        { name: "w", value: "400" }
      ],
      {
        pipeline: {
          steps: [
            {
              type: "rotate",
              mode: {
                type: "manual",
                degrees: 90,
                backgroundColor: "#00000000"
              }
            },
            {
              type: "rotate",
              mode: {
                type: "automatic"
              }
            },
            {
              type: "resize",
              geometry: {
                size: {
                  type: "widthxheight!",
                  width: 400,
                  height: 300
                }
              }
            }
          ]
        }
      }
    );
  });

  test("handle repetitions of same element type (separated by another element type)", async (): Promise<void> => {
    await assertTwoWayMap(
      imagePluginSchema as JSONSchema7,
      [
        { name: "rotate", value: "90" },
        { name: "w", value: "400" },
        { name: "h", value: "300" },
        { name: "fit", value: "stretch" },
        { name: "rotate", value: "180" }
      ],
      [
        { name: "rotate", value: "90" },
        { name: "rotate-bg", value: "00000000" },
        { name: "fit", value: "stretch" },
        { name: "h", value: "300" },
        { name: "w", value: "400" },
        { name: "rotate", value: "180" },
        { name: "rotate-bg", value: "00000000" }
      ],
      {
        pipeline: {
          steps: [
            {
              type: "rotate",
              mode: {
                type: "manual",
                degrees: 90,
                backgroundColor: "#00000000"
              }
            },
            {
              type: "resize",
              geometry: {
                size: {
                  type: "widthxheight!",
                  width: 400,
                  height: 300
                }
              }
            },
            {
              type: "rotate",
              mode: {
                type: "manual",
                degrees: 180,
                backgroundColor: "#00000000"
              }
            }
          ]
        }
      }
    );
  });

  async function assertTwoWayMap(
    schema: JSONSchema7,
    expectedQueryShorthand: OrderSensitiveQueryParams,
    expectedQueryCanonical: OrderSensitiveQueryParams,
    expectedJson: JsonObjectUnboxed
  ): Promise<void> {
    const mapper = await QueryStringToJsonMapper.fromJsonSchema(schema);

    // Test QueryString -> JSON
    const actualJsonFromShorthand = mapper.fromQuerystring(expectedQueryShorthand);
    expect(actualJsonFromShorthand.result).toStrictEqual(expectedJson);
    expect(actualJsonFromShorthand.invalidParamCombinations).toStrictEqual([]);
    expect(actualJsonFromShorthand.invalidParamTypes).toStrictEqual([]);
    expect(actualJsonFromShorthand.unknownParamNames).toStrictEqual([]);

    // Test QueryString -> JSON -> QueryString
    const actualQueryCanonical = mapper.toQuerystring(expectedJson);
    expect(actualQueryCanonical).toStrictEqual(expectedQueryCanonical);

    // Test QueryString -> JSON -> QueryString -> JSON
    const actualJsonFromCanonical = mapper.fromQuerystring(expectedQueryCanonical);
    expect(actualJsonFromCanonical.result).toStrictEqual(expectedJson);
    expect(actualJsonFromCanonical.invalidParamCombinations).toStrictEqual([]);
    expect(actualJsonFromCanonical.invalidParamTypes).toStrictEqual([]);
    expect(actualJsonFromCanonical.unknownParamNames).toStrictEqual([]);
  }
});
